/*
 * Slide-show server.
 *
 * Copyright 2007 by Antoine Mine'
 */


#include <sys/socket.h>
#include <arpa/inet.h>  /* Internet sockets */
#include <sys/un.h>     /* Unix sockets */
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/select.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <assert.h>

#include "protocol.h"
#include "pthread.h"

#include <SDL.h>
#include <SDL_video.h>
#include <SDL_keysym.h>
#include <SDL_image.h>



/*********************************************************************
 * SERVER PARAMETERS
 *********************************************************************/

/* Connection timeout, in seconds */
#define TIMEOUT 20

/* delai between frames, in ms */
#define FRAME_DELAY 100


#undef USE_OPENGL /* not implemented */

int debug = 0;


/*********************************************************************
 * SYSTEM UTILITIES
 *********************************************************************/

int my_endianess;

/* detect endianess */
void detect_endianess(void)
{
  uint8_t a[4] = { 0x01, 0x02, 0x03, 0x04 };
  uint32_t i= *(uint32_t*)a;
  if (i==0x01020304) {
    if (debug) fprintf(stderr, "big endian detected\n");
    my_endianess = ENDIANESS_BIG;
  }
  else if (i==0x04030201) {
    if (debug) fprintf(stderr, "little endian detected\n");
    my_endianess = ENDIANESS_LITTLE;
  }
  else {
    fprintf(stderr, "Unknown byte ordering (%08x).\n",i );
    exit(1);
  }
}

/* prints an error message, errno & quits */
void fatal_error(const char* msg)
{
  fprintf(stderr, "%s (errno: %s)\n", msg, strerror(errno) );
  exit(1);
}

/* wait for fd to be ready for read (rw=0) or write (rw=1) & return 0
   fails (return 1) after a time-out of t secs (0=no wait, <0=infnite)
*/
int timed_wait(int fd, int t, int rw)
{
  struct timeval to;
  fd_set f;
  while (1) {
    int r;
    FD_ZERO(&f);
    FD_SET(fd, &f);
    to.tv_sec = t;
    to.tv_usec = 0;
    r = select(fd+1, rw ? NULL : &f, rw ? &f : NULL, NULL, 
	       t>=0 ? &to : NULL); 
    if (r==-1) {
      if (errno==EINTR) continue; /* this may cause us to wait too long... */
      fprintf(stderr, "select error: %s\n", strerror(errno));
      return 1; /* error */
    }
    if (r==1 && FD_ISSET(fd, &f)) return 0;
    return 1; /* time-out */
  }
}

/* reads exctly count bytes from fd,
   returns 0 if OK, 1 if failure (e.g. file closed before count was read)
*/
int full_read(int fd, void* buf, size_t count)
{
  size_t i = 0;
  while (i<count) {
    ssize_t x;
    if (timed_wait(fd, TIMEOUT, 0)) return 1; /* timeout */
    x = read(fd, (char*)buf+i, count-i);
    if (x==-1) {
      if (errno==EINTR) continue;
      fprintf(stderr, "read error: %s\n", strerror(errno));
      return 1; /* error */
    }
    if (x==0) return 1; /* premature end of file */
    i += x;
  }
  return 0; /* OK */
}

/* writes exactly count bytes to fd,
   returns 0 if OK, 1 if failure (e.g. file closed before count was written)
*/
int full_write(int fd, void* buf, size_t count)
{
  size_t i = 0;
  while (i<count) {
    ssize_t x;
    if (timed_wait(fd, TIMEOUT, 1)) return 1; /* timeout */
    x = write(fd, (char*)buf+i, count-i);
    if (x==-1) {
      if (errno==EINTR) continue;
      if (errno==EPIPE) return 1;/* premature end of file */
      fprintf(stderr, "write error: %s\n", strerror(errno));
      return 1; /* error */
    }
    i += x;
  }
  return 0; /* OK */
}

/* skips count bytes in fd (i.e., read & discard)
   returns 0 if OK, 1 if failure (e.g. file closed before count was read)
*/
int skip(int fd, size_t count)
{
  char buf[512];
  while (count) {
    ssize_t x;
    if (timed_wait(fd, TIMEOUT, 0)) return 1; /* timeout */
    x = read(fd, buf, count<sizeof(buf) ? count : sizeof(buf));
    if (x==-1) {
      if (errno==EINTR) continue;
      fprintf(stderr, "read error: %s\n", strerror(errno));
      return 1; /* error */
    }
    if (x==0) return 1; /* premature end of file */
    count -= x;    
  }
  return 0;
}


/*********************************************************************
 * VIDEO HANDLING
 *********************************************************************/

#ifdef USE_OPENGL
#include <SDL_opengl.h>
#include <GL/gl.h>
#endif

SDL_Surface* screen = NULL;
SDL_Surface* images[65536];
int screen_width, screen_height, screen_flags;

#ifdef USE_OPENGL
volatile GLuint textures[65536];
#endif

volatile int shown = -1;
volatile double delay = 0;
volatile double elapsed = 0;
volatile int transition = 0;

/* protects all shared variables */
pthread_mutex_t mutex;

/* 'clean' signal handling */
void sig_int(int dummy)
{
  (void)dummy; /* avoids 'unused variable' warning */
  exit(1);     /* will call SDL_Quit through atexit */
}

void lock(void)
{
  errno = pthread_mutex_lock(&mutex);
  if (errno) fatal_error("pthread_mutex_lock failed");
}

void unlock(void)
{
  errno = pthread_mutex_unlock(&mutex);
  if (errno) fatal_error("pthread_mutex_unlock failed");
}


/* update screen size & flags, if necessary */
int screen_reinit(int width, int height, int flags)
{
  int f = SDL_DOUBLEBUF | SDL_HWSURFACE | SDL_ANYFORMAT;
#ifdef USE_OPENGL
  f |= SDL_OPENGL;
#endif
  if (screen && width==screen_width && height==screen_height && 
      flags==screen_flags) return 0; /* nothing to do */
  if (flags & SETUP_FULLSCREEN) f |= SDL_FULLSCREEN;
  screen_width  = width;
  screen_height = height;
  screen_flags  = flags;
  screen = SDL_SetVideoMode(screen_width, screen_height, 0, f);
  if (!screen) return 1;
  return 0;
}

/* called once, when server starts */
void video_init(void)
{
#ifdef USE_OPENGL
  /* OpenGL parameters */
  int gl_attr[][2] = { 
    { SDL_GL_RED_SIZE,         1 },
    { SDL_GL_GREEN_SIZE,       1 },
    { SDL_GL_BLUE_SIZE,        1 },
    { SDL_GL_ALPHA_SIZE,       0 },
    { SDL_GL_ACCUM_RED_SIZE,   0 },
    { SDL_GL_ACCUM_GREEN_SIZE, 0 },
    { SDL_GL_ACCUM_BLUE_SIZE,  0 },
    { SDL_GL_ACCUM_ALPHA_SIZE, 0 },
    { SDL_GL_DEPTH_SIZE,       0 },
    { SDL_GL_STENCIL_SIZE,     0 },
    { SDL_GL_STEREO,           0 },
    { SDL_GL_DOUBLEBUFFER,     1 },
  };
#endif
  unsigned i;
  if (SDL_Init(SDL_INIT_TIMER | SDL_INIT_VIDEO)) {
    fprintf(stderr, "SDL_Init failure: %s\n", SDL_GetError());
    exit(1);
  }

  atexit(SDL_Quit);
  signal(SIGHUP,sig_int);
  signal(SIGINT,sig_int);
  signal(SIGQUIT,sig_int);

  errno = pthread_mutex_init(&mutex, NULL);
  if (errno) fatal_error("pthread_mutex_init failed");

#ifdef USE_OPENGL
  /* OpenGL parameters */
  for (i=0;i<sizeof(gl_attr)/sizeof(gl_attr[0]);i++)
    if (SDL_GL_SetAttribute( gl_attr[i][0], gl_attr[i][1] ))
      fprintf(stderr, "SDL_SetAttribute(%i,%i) failure: %s\n", 
	      gl_attr[i][0], gl_attr[i][1], SDL_GetError());
#endif

  if (screen_reinit(320,240,0)) {
    fprintf(stderr, "SDL_SetVideoMode failure: %s\n", SDL_GetError());
    exit(1);
  }

  for (i=0;i<65536;i++) images[i] = NULL;
#ifdef USE_OPENGL
  for (i=0;i<65536;i++) textures[i] = 0;
#endif
}

/* returns the smallest power of 2 that is greater than x */
int round_pw2(int x)
{
  int y = 1;
  while (y<x) y *= 2;
  return y;
}

/* lazy texture upload to OpenGL */
void use_texture(int id)
{
#ifdef USE_OPENGL
  if (!textures[id]) {
    /* alloc texture */
    glGenTextures(1, &textures[id]);
    glBindTexture(GL_TEXTURE_2D, textures[id]);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 
		 round_pw2(images[id]->w), round_pw2(images[id]->h), 0,
		 GL_RGB, GL_UNSIGNED_BYTE, NULL);

    /* set parameters */
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    /* upload image */
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, images[id]->w, images[id]->h,
		    GL_RGB, GL_UNSIGNED_BYTE, images[id]->pixels);
  }
  else {
    /* texture already uploaded */
    glBindTexture(GL_TEXTURE_2D, textures[id]);
  }
#endif
}

/* find image to be shown after id, or -1 */
int next_id(int id)
{
  int i;

  if (id==-1) return -1;

  for (i=id+1; i<65536; i++)
    if (images[i]) return i;

  /* wrap-around */
  for (i=0;i<=id;i++)
    if (images[i]) return i;

  /* no more images :( */
  return -1;
}

/* display current frame */
void video_frame(void)
{
  int next;
  if (shown==-1 || !images[shown]) return;

  SDL_SetAlpha(images[shown], 0, SDL_ALPHA_OPAQUE);

  if (delay && elapsed>=delay-1 && transition!=TRANS_NONE &&
      (next=next_id(shown))!=-1 && next!=shown) {

    SDL_Rect p1 = {0,0} , p2 = {0,0}, p = {0,0,screen->w,screen->h};
    float f1 = elapsed-delay+1; /* 0->1 */
    float f2 = delay-elapsed;   /* 1->0 */


    switch (transition) {

    case TRANS_FADE:
      SDL_SetAlpha(images[shown], SDL_SRCALPHA, 
		   SDL_ALPHA_OPAQUE*f2+SDL_ALPHA_TRANSPARENT*f1);
      SDL_SetAlpha(images[next], SDL_SRCALPHA, 
		   SDL_ALPHA_OPAQUE*f1+SDL_ALPHA_TRANSPARENT*f2);
      SDL_BlitSurface(images[shown], NULL, screen, NULL);
      SDL_BlitSurface(images[next], NULL, screen, NULL);
      break;

    case TRANS_LEFT:
      p1.x = -screen->w*f1;
      p2.x = screen->w*f2;
      SDL_BlitSurface(images[shown], &p, screen, &p1);
      SDL_BlitSurface(images[next], &p, screen, &p2);
      break;

    case TRANS_RIGHT:
      p1.x = screen->w*f1;
      p2.x = -screen->w*f2;
      SDL_BlitSurface(images[shown], &p, screen, &p1);
      SDL_BlitSurface(images[next], &p, screen, &p2);
      break;

    case TRANS_UP:
      p1.y = -screen->w*f1;
      p2.y = screen->w*f2;
      SDL_BlitSurface(images[shown], &p, screen, &p1);
      SDL_BlitSurface(images[next], &p, screen, &p2);
      break;

    case TRANS_DOWN:
      p1.y = screen->w*f1;
      p2.y = -screen->w*f2;
      SDL_BlitSurface(images[shown], &p, screen, &p1);
      SDL_BlitSurface(images[next], &p, screen, &p2);
      break;
      

    default: 
      SDL_BlitSurface(images[shown], NULL, screen, NULL);
    }
  }
  else SDL_BlitSurface(images[shown], NULL, screen, NULL);
}


void video_loop(void)
{
  while (1) {
    SDL_Event event;

    /* wait for next tick */
    SDL_Delay(FRAME_DELAY);

    /* eats all events */
    while(SDL_PollEvent(&event)) {
      switch(event.type) {

	/* quit */
      case SDL_QUIT: exit(0);

      case SDL_KEYDOWN:

	/* escape => quit */
	if (event.key.keysym.sym==SDLK_ESCAPE) exit(0);

	/* ALT+ENTER => toggle fullscreen */
	if (event.key.keysym.sym==SDLK_RETURN && 
	    event.key.keysym.mod & KMOD_ALT)
	  screen_reinit(screen_width, screen_height, 
			screen_flags ^ SETUP_FULLSCREEN);
	break;
      }
    }

    lock();

    /* show current frame */
    SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format,0,0,0));
    video_frame();
    SDL_Flip(screen);

    /* time elapse */
    elapsed += FRAME_DELAY / 1000.;
    if (delay && elapsed >= delay) {
      elapsed = 0;
      shown = next_id(shown);
    }

    unlock();
  }
}


/*********************************************************************
 * PROTOCOL HANDLING
 *********************************************************************/

/* thread-local connection state */
typedef struct {
  int fd;         /* socket */
  int reply;      /* 1=server should reply, 0=should not (EPIPE received) */
  pthread_t id;   /* thread id */
  int endianess;  /* endianess */
} conn;


/* endianess conversions */

uint16_t conv16(conn* c, uint16_t i)
{
  if (my_endianess==c->endianess) return i;
  return (i>>8) | (i<<8);
}

uint32_t conv32(conn* c, uint32_t i)
{
  if (my_endianess==c->endianess) return i;
  return (i>>24) | ((i>>8) & 0xff00) | ((i<<8) & 0xff0000) | (i<<24);
}


/* replies */

void reply_id(conn* c, int type, int id)
{
  ans a = { type, conv16(c, id) };
  if (!c->reply) return;
  if (debug)
    fprintf(stderr, "%i: Reply %i, id=%u\n", (int)c->id, type, id);
  if (full_write(c->fd, &a, sizeof(a))) c->reply = 0; /* stop replying */
}

void reply(conn* c, int type)
{
  ans a = { type, 0 };
  if (!c->reply) return;
  if (debug)
    fprintf(stderr, "%i: Reply %i\n", (int)c->id, type);
  if (full_write(c->fd, &a, sizeof(a))) c->reply = 0; /* stop replying */
}

void handle_setup(conn* c, reqSetup* r)
{
  unsigned w = conv16(c, r->width);
  unsigned h = conv16(c, r->height);
  unsigned flags = r->flags;
  if (debug)
    fprintf(stderr, "%i: Got Setup request width=%u height=%u flags=%02X\n", 
	    (int)c->id, w, h, flags );
  lock();
  if (w>=2048 || h>=2048) {
    if (debug)
      fprintf(stderr, "%i: Dimension overflow\n", (int)c->id);
    unlock();
    reply(c, ANS_INVALID_ARGUMENT);
    return;
  }
  if (screen_reinit(w, h, flags)) {
    /* failure => try default values */
    if (debug) 
      fprintf(stderr, "%i: Setup failed, using 320x200 instead\n", (int)c->id);
    if (screen_reinit(320, 240, 0)) {
      /* we can do nothing more */
      fprintf(stderr, "SDL_SetVideoMode failure: %s\n", SDL_GetError());
      reply(c, ANS_INVALID_ARGUMENT);
      exit(1);
    }
    unlock();
    reply(c, ANS_INVALID_ARGUMENT);
    return;
  }
  unlock();
  reply(c, ANS_OK);
}

void def_image(conn* c, SDL_RWops* rw, int format, int id)
{
  SDL_Surface* s;
  switch (format) {
  case IMAGE_PNG: s = IMG_LoadPNG_RW(rw); break;
  case IMAGE_GIF: s = IMG_LoadGIF_RW(rw); break;
  case IMAGE_JPG: s = IMG_LoadJPG_RW(rw); break;
  case IMAGE_BMP: s = IMG_LoadBMP_RW(rw); break;
  default: 
    if (debug) 
      fprintf(stderr, "%i: Unknown image format %i\n", (int)c->id, format);
    reply(c, ANS_INVALID_ARGUMENT); return;
  }
  if (!s) {
    if (debug) fprintf(stderr, "%i: IMG_LoadXXX_RW failed\n", (int)c->id);
    reply(c, ANS_INVALID_IMAGE); 
    return;
  }
  reply(c, ANS_OK);
  lock();
  if (images[id]) SDL_FreeSurface(images[id]);
  images[id] = s;
#ifdef USE_OPENGL
  if (textures[id]) {
    glBindTexture(GL_TEXTURE_2D, 0);
    glDeleteTextures(1, & textures[id]);
  }
  textures[id] = 0;
#endif
  unlock();
}

void handle_def_image(conn* c, reqDefImage* r)
{
  unsigned id = conv16(c, r->id);
  unsigned size = conv32(c, r->size);
  SDL_RWops* rw = SDL_RWFromMem(r+1, size - sizeof(reqDefImage));
  if (debug)
    fprintf(stderr, "%i: Got DefImage request id=%u format=%u size=%u\n", 
	    (int)c->id, id, r->format, size);
  def_image(c, rw, r->format, id);
  SDL_FreeRW(rw);
}

void handle_def_image_SHM(conn* c, reqDefImageSHM* r)
{
  unsigned id = conv16(c, r->id);
  unsigned shmsize = conv32(c, r->shmsize);
  int shm = conv32(c, r->shm);
  SDL_RWops* rw;
  struct shmid_ds ds;
  void* ptr;

  if (debug)
    fprintf(stderr, 
	    "%i: Got DefImageSHM request id=%u format=%u shmsize=%u shm=%i\n", 
	    (int)c->id, id, r->format, shmsize, shm);

  /* check size */
  if (shmctl(shm, IPC_STAT, &ds)==-1) {
    if (debug) fprintf(stderr, "%i: shmctl failed\n", (int)c->id);
    reply(c, ANS_INVALID_SHM);
    return;
  }
  if (ds.shm_segsz < shmsize) {
    if (debug) fprintf(stderr, "%i: Invalid SHM size\n", (int)c->id);
    reply(c, ANS_INVALID_SHM);
    return;
  }
  /* get pointer */
  ptr = shmat(shm, NULL, 0);
  if (ptr == (void *) -1) {
    if (debug) fprintf(stderr, "%i: shmat failed\n", (int)c->id);
    reply(c, ANS_INVALID_SHM);
    return;
  }
  rw = SDL_RWFromMem(ptr, shmsize);
  def_image(c, rw, r->format, id);

  /* clean-up */
  SDL_FreeRW(rw);
  shmdt(ptr);
}

void handle_undef_image(conn* c, reqUndefImage* r)
{
  unsigned id = conv16(c, r->id);
  if (debug)
    fprintf(stderr, "%i: Got UndefImageSHM request id=%u\n", (int)c->id, id);
  if (id==0xffff) {
    int i;
    lock();
    for (i=0;i<0xffff;i++)
      if (images[i]) {
	SDL_FreeSurface(images[i]);
	images[i] = NULL;
      }
#ifdef USE_OPENGL
    glBindTexture(GL_TEXTURE_2D, 0);
    for (i=0;i<0xffff;i++)
      if (textures[i]) {
	glDeleteTextures(1, & textures[i]);
	textures[i] = 0;
      }
#endif 
    shown = -1;
    unlock();
    reply(c, ANS_OK);
  }
  else if (images[id]) {
    lock();
    SDL_FreeSurface(images[id]);
    images[id] = NULL;
#ifdef USE_OPENGL
    glBindTexture(GL_TEXTURE_2D, 0);
    if (textures[id]) {
      glDeleteTextures(1, & textures[id]);
      textures[id] = 0;
    }
#endif
    if (id==shown) 
      shown = next_id(shown);
    unlock();
    reply(c, ANS_OK);
  }
  else { 
    if (debug)
      fprintf(stderr, "%i: Unknown identifier %i\n", (int)c->id, id);
    reply(c, ANS_UNKNOWN_IDENTIFIER);
  }
}

void handle_show_image(conn* c, reqShowImage* r)
{
  unsigned id = conv16(c, r->id);
  if (debug)
    fprintf(stderr, "%i: Got ShowImage request id=%u\n", (int)c->id, id);
  lock();
  if (!images[id]) {
    if (debug)
      fprintf(stderr, "%i: Unknown identifier %i\n", (int)c->id, id);
    unlock();
    reply(c, ANS_UNKNOWN_IDENTIFIER);
    return;
  }
  shown = id;
  transition = r->trans;
  delay = conv16(c, r->delay);
  elapsed = 0;
  unlock();
  reply(c, ANS_OK);
}

void handle_query_image(conn* c, reqQueryImage* r)
{
  int i;
  unsigned id = conv16(c, r->id);
  if (debug)
    fprintf(stderr, "%i: Got QueryImage request query=%u id=%i\n", 
	    (int)c->id, r->query, id);

  switch (r->query) {

  case QUERY_FIRST:
    lock();
    for (i=0;i<65536;i++)
      if (images[i]) {
	unlock();
	reply_id(c, ANS_OK, i);
	return;
      }
    unlock();
    if (debug)
      fprintf(stderr, "%i: Empty image list\n", (int)c->id);
    reply(c, ANS_UNKNOWN_IDENTIFIER);
    break;

  case QUERY_LAST:
    lock();
    for (i=65535;i>=0;i--)
      if (images[i]) {
	unlock();
	reply_id(c, ANS_OK, i);
	return;
      }
    unlock();
    if (debug)
      fprintf(stderr, "%i: Empty image list\n", (int)c->id);
    reply(c, ANS_UNKNOWN_IDENTIFIER);
    break;

  case QUERY_SHOWN:
    lock();
    if (shown>=0) { unlock(); reply_id(c, ANS_OK, shown); }
    else { 
      unlock(); 
      if (debug)
	fprintf(stderr, "%i: No image shown\n", (int)c->id);
      reply(c, ANS_UNKNOWN_IDENTIFIER); 
    }
    break;

  case QUERY_NEXT:
    lock();
    for (i=id+1;i<65536;i++)
      if (images[i]) {
	unlock();
	reply_id(c, ANS_OK, i);
	return;
      }
    unlock();
    if (debug)
      fprintf(stderr, "%i: No next image\n", (int)c->id);
    reply(c, ANS_UNKNOWN_IDENTIFIER);
    break;

  case QUERY_UNUSED:
    lock();
    for (i=0;i<65536;i++)
      if (!images[i]) {
	unlock();
	reply_id(c, ANS_OK, i);
	return;
      }
    unlock();
    if (debug)
      fprintf(stderr, "%i: Image list full\n", (int)c->id);
    reply(c, ANS_UNKNOWN_IDENTIFIER);
    break;

  default:
    if (debug)
      fprintf(stderr, "%i: Unknowk query type %i\n", (int)c->id, r->query);
    reply(c, ANS_INVALID_ARGUMENT);
  }
}

void handle_set_endianess(conn* c, reqSetEndianess* r)
{
  if (debug)
    fprintf(stderr, "%i: Got SetEndianess request endianess=%i\n", 
	    (int)c->id, r->endianess);
  if (r->endianess==ENDIANESS_LITTLE || 
      r->endianess==ENDIANESS_BIG) {
    c->endianess = r->endianess;
    reply(c, ANS_OK);
  }
  else {
   if (debug)
      fprintf(stderr, "%i: Unknown endianess type %i\n", 
	      (int)c->id, r->endianess);
     reply(c, ANS_INVALID_ARGUMENT);
  }
}

int dispatch_request(conn* c, void* ptr, unsigned size)
{
  req* r = (req*) ptr;

#define HANDLE(req,struct,fun)			\
  case req:								\
    if (size==sizeof(struct)) fun(c, (struct*)ptr);			\
    else {								\
      if (debug)							\
	fprintf(stderr, "%i: Invalid size for request\n", (int)c->id);	\
      reply(c, ANS_INVALID_REQUEST);					\
    }								\
  break							      
  
  /* dispatch */
  switch (r->type) {

    HANDLE(SETUP_REQ, reqSetup, handle_setup);
    HANDLE(UNDEF_IMAGE_REQ, reqUndefImage, handle_undef_image);
    HANDLE(SHOW_IMAGE_REQ, reqShowImage, handle_show_image);
    HANDLE(QUERY_IMAGE_REQ, reqQueryImage, handle_query_image);
    HANDLE(DEF_IMAGE_SHM_REQ, reqDefImageSHM, handle_def_image_SHM);
    HANDLE(SET_ENDIANESS_REQ, reqSetEndianess, handle_set_endianess);

  case DEF_IMAGE_REQ:
    handle_def_image(c, (reqDefImage*)ptr);
    break;

  default:
    fprintf(stderr, "%i: Unknown request type %i\n", (int)c->id, r->type);
    reply(c, ANS_INVALID_REQUEST);
  }

  return 0;
}

/* return 1 if the connection is closed */
int handle_request(conn* c)
{
  int x;
  unsigned size;
  char* ptr;
  req r;
  
  /* read common header */
  if (full_read(c->fd, &r, sizeof(r))) return 1;
  size = conv32(c, r.size);

  /* special case: size overriden */
  if (r.type==SET_ENDIANESS_REQ) size = sizeof(reqSetEndianess);

  if (debug)
    fprintf(stderr, "%i: Got request header type=%u size=%u\n", 
	    (int)c->id, r.type, size);
  
  /* alloc buffer to hold query */
  if (size>=(1<<25) || (ptr=malloc(size))==NULL) {
    if (debug)
      fprintf(stderr, "%i: Request too big, aborting\n", (int)c->id);
    reply(c, ANS_INVALID_REQUEST);
    if (skip(c->fd, size-sizeof(r))) return 1;
    return 0;
  }

  /* read remaining of the query */
  memcpy(ptr, &r, sizeof(r));
  if (full_read(c->fd, ptr+sizeof(r), size-sizeof(r))) {
    free(ptr);
    if (debug)
      fprintf(stderr, "%i: Unfinished request\n", (int)c->id);
    return 1;
  }

  x = dispatch_request(c, ptr, size);
  free(ptr);
  return x;
}


/*********************************************************************
 * UNIX SERVER
 *********************************************************************/

int stream_endianess;

/* handle whole connection; called in a separate thread */
void* handle_connection(void* param)
{
  conn c;
  c.fd = (int)param;
  c.reply = 1;
  c.id = pthread_self();
  c.endianess = stream_endianess;
  if (debug)
    fprintf(stderr, "%i: New thread created\n", (int)c.id);
  while (1) {
    if (handle_request(&c)) break;
  }
  close(c.fd);
  if (debug)
    fprintf(stderr, "%i: Connection closed\n", (int)c.id);
  return NULL;
}

/* create & return a Unix stream domain socket to ADDR_UNIX */
int create_unix_socket(void)
{
  int listen_fd;           /* Unix socket */
  struct sockaddr_un addr; /* Unix socket address */

  /* create a Unix stream socket */
  listen_fd = socket(PF_UNIX, SOCK_STREAM, 0);
  if (listen_fd==-1) fatal_error("Unix socket creation failed");

  /* fills addr with the socket filename ADDR_UNIX */
  assert(strlen(ADDR_UNIX)<sizeof(addr.sun_path));
  memset(&addr, 0, sizeof(addr));    /* 0-initialized */
  addr.sun_family = AF_UNIX;        /* Unix socket */
  strcpy(addr.sun_path, ADDR_UNIX);  /* filename */
  
  /* bind fails if ADDR_UNIX already exists */
  unlink(ADDR_UNIX);

  /* binds socket to ADDR_UNIX */
  if (bind(listen_fd, (const struct sockaddr*) &addr, sizeof(addr))==-1)
    fatal_error("Unix socket bind failed");
 
  /* makes liste_fd a listening socket */
  if (listen(listen_fd, 15)==-1) fatal_error("Unix socket listen failed");
  
  if (debug)
    fprintf(stderr, "Unix socket server listening to %s\n", ADDR_UNIX);

  return listen_fd;
}

/* loop managing Unix connections to listen_fd: called in a separate thread */
void* unix_accept_loop(void* param)
{
  int listen_fd = (int)param;
  while (1) {
    pthread_t thread;

    /* waits for connection */
    int fd = accept(listen_fd, NULL, 0);
    if (fd==-1) {
      if (errno==EINTR) continue; /* non-fatals */
      fatal_error("Unix socket accept failure");
    }

    if (debug) fprintf(stderr, "Unix connection accepted\n");

    /* create new thread */
    errno = pthread_create(&thread, NULL, handle_connection, (void*)fd);
    if (errno) {
      fprintf(stderr, "pthread_create failed (%s)\n", strerror(errno));
      close(fd);
    }
    else pthread_detach(thread);
  }

  return NULL;
}

/* starts the Unix accept loop in separate thread */
void start_unix_server(void)
{
  pthread_t thread;
  int fd;
  fd = create_unix_socket();
  errno = pthread_create(&thread, NULL, unix_accept_loop, (void*)fd);
  if (errno) {
    fprintf(stderr, "pthread_create failed (%s)\n", strerror(errno));
    exit(1);
  }
}

/*********************************************************************
 * INTERNET TCP SERVER
 *********************************************************************/

/* create & return an Internet TCP listening socket */
int create_tcp_socket(void)
{
  int listen_fd;           /* listening socket */
  struct sockaddr_in addr; /* IPv4 listening addr */
  int one = 1;             /* used with setsockopt */
  
  /* creates a TCP internet socket */
  listen_fd = socket(PF_INET, SOCK_STREAM, 0);
  if (listen_fd==-1) fatal_error("Internet TCP socket creation failed");
  
  /* removes the delay between two binds of the same port */
  if (setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one))==-1)
    fatal_error("setsockopt(SO_REUSEADDR) failed");

  /* fills addr with port */
  memset(&addr, 0, sizeof(addr));   /* 0-initialization: MANDATORY */
  addr.sin_family = AF_INET;        /* Internet socket */
  addr.sin_port   = htons(PORT_IP); /* port: 2 bytes in network order */
  addr.sin_addr.s_addr = htonl(INADDR_ANY); /* accepts on all interfaces */

  /* binds listen_fd to addr */
  if (bind(listen_fd, (const struct sockaddr*) &addr, sizeof(addr))==-1)
    fatal_error("Internet TCP socket bind failed");

  /* makes liste_fd a listening socket */
  if (listen(listen_fd, 15)==-1) 
    fatal_error("Internet TCP socket listen failed");
  
  if (debug) 
    fprintf(stderr, "Internet TCP socket server listening on port %i\n", PORT_IP);

  return listen_fd;
}


/* loop managing TCP connections to listen_fd: called in a separate thread */
void* tcp_accept_loop(void* param)
{
  int listen_fd = (int)param;
  while (1) {
    struct sockaddr_in addr;            /* client IP address */
    socklen_t addr_len = sizeof(addr);  /* IPv4 address size */
    char buf[256];                      /* to store addr as string */
    pthread_t thread;

    /* waits for connection */
    int fd = accept(listen_fd, (struct sockaddr*)&addr, &addr_len);
    if (fd==-1) {
      if (errno==EINTR || errno==ECONNABORTED) continue; /* non-fatals */
      fatal_error("Internet TCP socket accept failure");
    }

    /* prints */
    if (debug) {
      if (inet_ntop(AF_INET, &addr.sin_addr, buf, sizeof(buf))) {
	fprintf(stderr, "Internet TCP connection accepted from %s:%i\n",
		buf, ntohs(addr.sin_port));
      }
      else {
	fprintf(stderr, "Internet TCP connection accepted\n");
      }
    }

    /* create new thread */
    errno = pthread_create(&thread, NULL, handle_connection, (void*)fd);
    if (errno) {
      fprintf(stderr, "pthread_create failed (%s)\n", strerror(errno));
      close(fd);
    }
    else pthread_detach(thread);
  }

  return NULL;
}

/* starts the TCP accept loop in separate thread */
void start_tcp_server(void)
{
  pthread_t thread;
  int fd;
  fd = create_tcp_socket();
  errno = pthread_create(&thread, NULL, tcp_accept_loop, (void*)fd);
  if (errno) {
    fprintf(stderr, "pthread_create failed (%s)\n", strerror(errno));
    exit(1);
  }
}


/*********************************************************************
 * INTERNET UDP SERVER
 *********************************************************************/

int packet_endianess;

/* create & return an Internet TCP listening socket */
int create_udp_socket(void)
{
  int fd;
  struct sockaddr_in addr; /* IPv4 listening addr */

  /* creates a UDP internet socket */
  fd = socket(PF_INET, SOCK_DGRAM, 0);
  if (fd==-1) fatal_error("Internet UDP socket creation failed");

  /* fills addr with port */
  memset(&addr, 0, sizeof(addr));   /* 0-initialization: MANDATORY */
  addr.sin_family = AF_INET;        /* Internet socket */
  addr.sin_port   = htons(PORT_IP); /* port: 2 bytes in network order */
  addr.sin_addr.s_addr = htonl(INADDR_ANY); /* accepts on all interfaces */

  /* binds fd to addr */
  if (bind(fd, (const struct sockaddr*) &addr, sizeof(addr))==-1)
    fatal_error("Internet UDP socket bind failed");

  if (debug)
    fprintf(stderr, "Internet UDP socket server listening on port %i\n", PORT_IP);

  return fd;
}

/* loop receiving UDP datagrams: called in a separate thread */
void* udp_loop(void* param)
{
  static char bigbuf[65536]; /* > to max length of UDP datagram */
  req* r = (req*)bigbuf;
  int fd = (int)param;

  while (1) {
    struct sockaddr_in addr;            /* client IP address */
    socklen_t addr_len = sizeof(addr);  /* IPv4 address size */
    char buf[256];                      /* to store addr as string */
    ssize_t s;
    conn c;

    c.fd = fd;
    c.reply = 1;
    c.id = pthread_self();
    c.endianess = packet_endianess;

   /* waits for datagram */
    s = recvfrom(fd, bigbuf, sizeof(bigbuf), 0, 
		 (struct sockaddr*)&addr, &addr_len);
    if (s==-1) {
      if (errno==EINTR) continue;
      fatal_error("Internet UDP socket recvfrom failure");
    }
    if (s==0) continue;

    /* prints */
    if (debug) {
      if (inet_ntop(AF_INET, &addr.sin_addr, buf, sizeof(buf))) {
	fprintf(stderr, "Internet UDP datagram from %s:%i\n",
		buf, ntohs(addr.sin_port));
      }
      else {
	fprintf(stderr, "Internet UDP datagram\n");
      }
    }

    /* connect to simplify reply sending */
    if (connect(fd, (const struct sockaddr*) &addr, addr_len)==-1) {
      fprintf(stderr, "Internet UDP connect failed\n");
      continue;
    }

    /* check request */
    if (s<sizeof(*r) ||	s<conv32(&c, r->size)) {
      if (debug)
	fprintf(stderr, "%i: Invalid datagram size: %i\n", (int)c.id, s);
      reply(&c, ANS_INVALID_REQUEST);
      continue;
    }
    if (r->type==SET_ENDIANESS_REQ)  {
      if (debug)
	fprintf(stderr, "%i: Invalid datagram request type: SetEndianess\n",
		(int)c.id);
      reply(&c, ANS_INVALID_REQUEST);
      continue;
    }

    /* handle datagram */
    dispatch_request(&c, bigbuf, s);
  }

  return NULL;
}

/* starts the UDP loop in separate thread */
void start_udp_server(void)
{
  pthread_t thread;
  int fd;
  fd = create_udp_socket();
  errno = pthread_create(&thread, NULL, udp_loop, (void*)fd);
  if (errno) {
    fprintf(stderr, "pthread_create failed (%s)\n", strerror(errno));
    exit(1);
  }
}

/*********************************************************************
 * HELP
 *********************************************************************/

void help(void)
{
  printf("options:\n  -h  help\n  -d  debug\n");
  printf("  -l  little-endian\n  -g  big-endian\n");
  exit(0);
}


/*********************************************************************
 * MAIN
 *********************************************************************/

int main(int argc, char* argv[])
{
  int i;

  /* default endianess */
  detect_endianess();
  stream_endianess = my_endianess;
  packet_endianess = ENDIANESS_BIG;

  /* options */
  for (i=1; i<argc; i++) {
    if (!strcmp(argv[i],"-d")) debug = !debug;
    else if (!strcmp(argv[i],"-b")) 
      stream_endianess = packet_endianess = ENDIANESS_BIG;
    else if (!strcmp(argv[i],"-l")) 
      stream_endianess = packet_endianess = ENDIANESS_LITTLE;
    else if (!strcmp(argv[i],"-h")) help();
    else {
      fprintf(stderr, "unknown option %s\n", argv[i]);
    }
  }

  /* start */
  video_init();
  signal(SIGPIPE,SIG_IGN);
  start_unix_server();
  start_tcp_server();
  start_udp_server();
  video_loop();
  return 0;
}
