/*
 * Definition of the 'slideshow' protocol.
 *
 * Copyright 2007 by Antoine Mine'
 */

/* avoids #including the same file twice */
#ifndef PROTOCOL_H_
#define PROTOCOL_H_


/* to get (u)int8/16/32_t */
#include <inttypes.h>
/*#include <stdint.h>*/


/*********************************************************************
 * CONFIGURATION
 *********************************************************************/

/* Unix domain socket name of server.
 */
#define ADDR_UNIX "/tmp/slide_server"


/* Internet TCP port of server.
   Must be > 1023 because the 1024 lower ports are reserved to root servers.
*/
#define PORT_IP  2010


/*********************************************************************
 * ENUMERATIONS
 *********************************************************************/


/* Flags for setupReq */
typedef enum {

  SETUP_FULLSCREEN = 1,

} setupFlags;


/* Image formats, see reqDefImage, reqDefImageSHM. */
typedef enum {

  IMAGE_PNG = 1, /* PNG file */
  IMAGE_GIF = 2, /* GIF file */
  IMAGE_JPG = 3, /* JPG file */
  IMAGE_BMP = 4, /* BMP file */

} imageFormat;


/* Transition between images, see reqShowImage. */
typedef enum {

  TRANS_NONE  = 1, /* immediate replace */
  TRANS_FADE  = 2, /* fade between images */
  TRANS_LEFT  = 3, /* images slide leftwards */
  TRANS_RIGHT = 4, /* images slide rightwards */
  TRANS_UP    = 5, /* images slide upwards */
  TRANS_DOWN  = 6, /* images slide downwards */

} transType;


/* Query types, see reqQueryImage */
typedef enum {

  QUERY_FIRST  = 1, /* first identifier associated to an image */
  QUERY_LAST   = 2, /* last identifier associated to an image */
  QUERY_SHOWN  = 3, /* identifier of the image currently show */
  QUERY_NEXT   = 4, /* identifier of the image following some base image */
  QUERY_UNUSED = 5, /* first unused identifier */

} queryType;


/* Endianess, see reqSetEndianess */
typedef enum {

  ENDIANESS_LITTLE = 1, /* less significant byte first */
  ENDIANESS_BIG    = 2, /* most significant byte first */

} endianess;


/*********************************************************************
 * REQUESTS
 *********************************************************************/


/* Request types. */
typedef enum {

  SETUP_REQ         = 1,
  DEF_IMAGE_REQ     = 2,
  UNDEF_IMAGE_REQ   = 3,
  SHOW_IMAGE_REQ    = 4,
  QUERY_IMAGE_REQ   = 5,
  DEF_IMAGE_SHM_REQ = 6,
  SET_ENDIANESS_REQ = 7,

} reqType;



/* Header common to all requests. */
typedef struct {

  uint8_t  type;  /* see reqType */
  uint32_t size;  /* size of message in bytes, including header and data */

} req;


/* Screen setup request. */
typedef struct {

  /* header */
  uint8_t  type;    /* = SETUP_REQ */
  uint32_t size;    /* = sizeof(reqSetup) */

  /* data */
  uint8_t  flags;   /* combination (|) of setupFlags */
  uint16_t width;
  uint16_t height;

} reqSetup;


/* Sends an image to a server and associate it to a client-chosen identifier. */
typedef struct {
  
  /* header */
  uint8_t  type;    /* = DEF_IMAGE_REQ */
  uint32_t size;    /* = sizeof(reqDefImage) + byte size of encoded image */

  /* data */
  uint16_t id;      /* client-chosen identifier to reference to image */
  uint8_t  format;  /* see imageFormat */
  
  /* encoded contents of image follows directly, variable length */

} reqDefImage;


/* Deletes an image in the server. */
typedef struct {

  /* header */
  uint8_t  type;    /* = UNDEF_IMAGE_REQ */
  uint32_t size;    /* = sizeof(reqUndefImage) */
  
  /* data */
  uint16_t id;      /* identifier of image to delete, 0xffff means all */

} reqUndefImage;


/* Starts displaying images. */
typedef struct {

  /* header */
  uint8_t  type;    /* = SHOW_IMAGE_REQ */
  uint32_t size;    /* = sizeof(reqShowImage) */

  /* data */
  uint16_t id;      /* identifier of first image to show */
  uint16_t delay;   /* delay between two images, 0 means infinite */
  uint8_t  trans;   /* see transType */
   
} reqShowImage;


/* Queries for the identifier of some image. */
typedef struct {
  
  /* header */
  uint8_t  type;    /* = QUERY_IMAGE_REQ */
  uint32_t size;    /* = sizeof(reqQueryImage) */

  /* data */
  uint16_t id;      /* base image identifier, used by QUERY_NEXT query */
  uint8_t  query;   /* query type, see queryType */

} reqQueryImage;


/* Sends an image through a shared memory segment. */
typedef struct {
  
  /* header */
  uint8_t  type;    /* = DEF_IMAGE_SHM_REQ */
  uint32_t size;    /* = sizeof(reqDefImageSHM) */

  /* data */
  uint32_t shm;     /* identifier of shared memory segment, as returned by shmget */
  uint32_t shmsize; /* size of shared memory segment in bytes */
  uint16_t id;      /* client-chosen identifier to reference to image */
  uint8_t  format;  /* see imageFormat */

} reqDefImageSHM;


/* Sets endianess */
typedef struct {
  
  /* header */
  uint8_t type;      /* = SET_ENDIANESS_REQ */
  uint32_t size;     /* ignored */

  /* data */
  uint8_t endianess; /* see endianess */
  
} reqSetEndianess;


/*********************************************************************
 * ANSWERS
 *********************************************************************/

/* Answer types. */
typedef enum {

  ANS_OK                 = 1,
  ANS_INTERNAL_ERROR     = 2, /* internal server error */
  ANS_INVALID_REQUEST    = 3, /* malformed request */
  ANS_INVALID_IMAGE      = 4, /* invalid image data in reqDefImage(SHM) */
  ANS_UNKNOWN_IDENTIFIER = 5, /* identifier is not associated to any image */
  ANS_INVALID_SHM        = 6, /* invalid SHM segment */
  ANS_INVALID_ARGUMENT   = 7, /* other invalid argument */

} ansType;



/* All answers have the same simple format. */
typedef struct {

  uint8_t  type;  /* see ansType */
  uint16_t data;  /* an image identifier, or meaningless */

} ans;



#endif /* PROTOCOL */

